package pwc.taxtech.atms.vat.service.impl;

import com.google.common.collect.Lists;
import org.apache.commons.lang3.StringUtils;
import org.apache.poi.openxml4j.exceptions.InvalidFormatException;
import org.apache.poi.ss.formula.functions.FreeRefFunction;
import org.apache.poi.ss.formula.udf.AggregatingUDFFinder;
import org.apache.poi.ss.formula.udf.DefaultUDFFinder;
import org.apache.poi.ss.formula.udf.UDFFinder;
import org.apache.poi.ss.usermodel.Cell;
import org.apache.poi.ss.usermodel.FormulaEvaluator;
import org.apache.poi.ss.usermodel.Row;
import org.apache.poi.ss.usermodel.Sheet;
import org.apache.poi.ss.usermodel.Workbook;
import org.apache.poi.ss.usermodel.WorkbookFactory;
import org.apache.poi.xssf.usermodel.XSSFCell;
import org.apache.poi.xssf.usermodel.XSSFWorkbook;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;
import pwc.taxtech.atms.common.POIUtil;
import pwc.taxtech.atms.common.util.SpringContextUtil;
import pwc.taxtech.atms.constant.enums.CellDataSourceType;
import pwc.taxtech.atms.constant.enums.FormulaDataSourceType;
import pwc.taxtech.atms.dao.ProjectMapper;
import pwc.taxtech.atms.dto.vatdto.CellCalcInfoDto;
import pwc.taxtech.atms.dto.vatdto.CellTemplateConfigGroupDto;
import pwc.taxtech.atms.entity.Project;
import pwc.taxtech.atms.vat.dpo.PeriodCellTemplateConfigExtendDto;
import pwc.taxtech.atms.vat.entity.*;
import pwc.taxtech.atms.vat.service.ReportGenerator;
import pwc.taxtech.atms.vat.service.impl.report.functions.*;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.stream.Collectors;

import static pwc.taxtech.atms.common.util.SpringContextUtil.reportMapper;
import static pwc.taxtech.atms.constant.Constant.EMPTY;

@Component
public class ReportGeneratorImpl extends VatAbstractService implements ReportGenerator {

    private static final Logger logger = LoggerFactory.getLogger(ReportGeneratorImpl.class);
    @Autowired
    private ProjectMapper projectMapper;

    @Autowired
    private FormulaAgent formulaAgent;

    @Override
    public String generateData(String projectId, List<Long> templateIds, Boolean ifDeleteManualDataSource,
                               List<String> exceptCellTemplates, Integer periodParam, Optional<String> generator) {
        Date createTime = new Date();
        int period = periodParam != null ? periodParam : 0;

        Project project = projectMapper.selectByPrimaryKey(projectId);
        if (templateIds.isEmpty()) templateIds.add(Long.MAX_VALUE);

        PeriodTemplateExample periodTemplateExample = new PeriodTemplateExample();
        periodTemplateExample.createCriteria().andProjectIdEqualTo(projectId).andPeriodEqualTo(period).andIdIn(templateIds);
        List<PeriodTemplate> periodTemplateList = periodTemplateMapper.selectByExample(periodTemplateExample);
        List<Long> periodTemplateIdList = periodTemplateList.stream()
                .map(PeriodTemplate::getTemplateId)
                .collect(Collectors.toList());

        Long templateGroupId = periodTemplateList.size() > 0 ? periodTemplateList.get(0).getTemplateGroupId() : 0;

        PeriodCellTemplateExample periodCellTemplateExample = new PeriodCellTemplateExample();
        if (periodTemplateIdList.isEmpty()) periodTemplateIdList.add(Long.MAX_VALUE);
        periodCellTemplateExample.createCriteria().andProjectIdEqualTo(projectId).andReportTemplateIdIn(periodTemplateIdList).andPeriodEqualTo(period);
        List<PeriodCellTemplate> periodCellTemplateList = periodCellTemplateMapper.selectByExample(periodCellTemplateExample);

        PeriodCellTemplateConfigExample periodCellTemplateConfigExample = new PeriodCellTemplateConfigExample();
        // only get formula config
        periodCellTemplateConfigExample.createCriteria().andProjectIdEqualTo(projectId).andReportTemplateIdIn(periodTemplateIdList).andPeriodEqualTo(period);
        List<PeriodCellTemplateConfig> periodCellTemplateConfigList =
                periodCellTemplateConfigMapper.selectByExample(periodCellTemplateConfigExample);

        //Map<String, String> templateCodeAndPath = new HashMap<>();
        //periodTemplateList.forEach(a -> templateCodeAndPath.put(a.getCode(), getTemplatePath(a.getPath())));
        Optional<Workbook> workbook = createWorkBookWithExcelFileList(periodTemplateList);

        if (workbook.isPresent()) {
            Workbook newWorkbook = workbook.get();
            FormulaContext formulaContext = new FormulaContext();
            formulaContext.setFormulaAgent(formulaAgent);
            formulaContext.setPeriod(period);
            formulaContext.setReportTemplateGroupId(templateGroupId);
            formulaContext.setProjectId(project.getId());
            formulaContext.setYear(project.getYear());
            formulaContext.setIsYear(period == 0);
            formulaContext.setIfRound(true);

            formulaContext.setOrganizationId(project.getOrganizationId());
            formulaContext.setIfRound(true);

            addFunctionsToWorkbook(newWorkbook, formulaContext);
            int sheetCount = newWorkbook.getNumberOfSheets();
            List<Cell> otherFormulaCells = new ArrayList<>();
            List<Cell> bbFormulaCells = new ArrayList<>();

            for (int i = 0; i < sheetCount; i++) {
                Sheet sheet = newWorkbook.getSheetAt(i);
                String code = sheet.getSheetName();
                Optional<PeriodTemplate> periodTemplate = periodTemplateList.stream()
                        .filter(a -> a.getCode().equals(code))
                        .findFirst();
                Long templateId;
                if (periodTemplate.isPresent()) {
                    templateId = periodTemplate.get().getTemplateId();
                } else {
                    templateId = 0L;
                }

                if (templateId > 0) {
                    //get cell template and cell template config with templateId
                    Map<PeriodCellTemplate, PeriodCellTemplateConfig> cellTemplatePeriodCellTemplateConfigMap = new HashMap<>();
                    List<PeriodCellTemplate> tempPeriodCellTemplateList = periodCellTemplateList.stream()
                            .filter(a -> a.getReportTemplateId().equals(templateId))
                            .collect(Collectors.toList());
                    tempPeriodCellTemplateList.forEach(a -> {
                        Optional<PeriodCellTemplateConfig> tempCellTemplateConfig = periodCellTemplateConfigList.stream()
                                .filter(item -> item.getCellTemplateId().equals(a.getCellTemplateId())
                                        && item.getDataSourceType().equals(CellDataSourceType.Formula.getCode()))
                                .findFirst();
                        tempCellTemplateConfig.ifPresent(periodCellTemplateConfig ->
                                cellTemplatePeriodCellTemplateConfigMap.put(a, periodCellTemplateConfig));
                    });

                    cellTemplatePeriodCellTemplateConfigMap.forEach((k, v) -> {
                        Row row = sheet.getRow(k.getRowIndex());
                        if (null == row) {
                            sheet.createRow(k.getRowIndex());
                        }
                        row = sheet.getRow(k.getRowIndex());

                        Cell cell = row.getCell(k.getColumnIndex());
                        if (null == cell) {
                            row.createCell(k.getColumnIndex());
                        }
                        cell = row.getCell(k.getColumnIndex());

                        //todo:后面单独处理kv的公式
                        if (StringUtils.isNotBlank(v.getFormula()) && !v.getFormula().contains("@")) {
                            cell.setCellFormula(v.getFormula());
                            logger.debug("formula:" + v.getFormula());
                            if (v.getFormula().startsWith("BB(")) {
                                bbFormulaCells.add(cell);
                            } else {
                                otherFormulaCells.add(cell);
                            }
                            //kv 公式处理
                        } else if (v.getFormula().contains("@")) {
                            if (StringUtils.isNotBlank(v.getKeyValueParsedFormula())) {
                                cell.setCellFormula(v.getKeyValueParsedFormula());
                                logger.debug("formula:" + v.getKeyValueParsedFormula());
                                otherFormulaCells.add(cell);
                            }
                        }
                    });
                }
            }
            FormulaEvaluator evaluator = newWorkbook.getCreationHelper().createFormulaEvaluator();
            otherFormulaCells.stream().forEach(m -> {
                evaluator.evaluateFormulaCellEnum(m);
            });
            bbFormulaCells.stream().forEach(n -> {
                evaluator.evaluateFormulaCellEnum(n);
            });

            //todo:loop all the cell which existing in the cellTemplateConfig, save the data to DB
            for (int i = 0; i < sheetCount; i++) {
                Sheet sheet = newWorkbook.getSheetAt(i);
                String code = sheet.getSheetName();
                Optional<PeriodTemplate> periodTemplate = periodTemplateList.stream()
                        .filter(a -> a.getCode().equals(code))
                        .findFirst();
                Long templateId;
                if (periodTemplate.isPresent()) {
                    templateId = periodTemplate.get().getTemplateId();
                } else {
                    templateId = 0L;
                }

                if (templateId > 0) {
                    //todo: create report into DB
                    PeriodReport report = new PeriodReport();
                    Long reportId = distributedIdService.nextId();
                    report.setId(reportId);
                    report.setTemplateId(templateId);
                    report.setPeriod(period);
                    report.setProjectId(projectId);
                    report.setCreateBy("Admin");
                    report.setCreateTime(createTime);
                    report.setUpdateBy("Admin");
                    report.setUpdateTime(createTime);
                    report.setProjectId(projectId);
                    reportMapper.insertSelective(report);

                    List<PeriodCellTemplateConfig> periodCellTemplateConfigs = periodCellTemplateConfigList.stream()
                            .filter(a -> a.getReportTemplateId().equals(templateId) && a.getDataSourceType().equals(CellDataSourceType.Formula.getCode()))
                            .collect(Collectors.toList());

                    List<PeriodCellTemplateConfig> keyInCellTemplateConfigs = periodCellTemplateConfigList.stream()
                            .filter(a -> a.getReportTemplateId().equals(templateId) && a.getDataSourceType().equals(CellDataSourceType.KeyIn.getCode()))
                            .collect(Collectors.toList());

                    //todo: add manual datasource here,in order to disaplay the manual datasource we need add it first at here.
//                    for (PeriodCellTemplateConfig keyInCellTemplateConfig : keyInCellTemplateConfigs) {
//                        PeriodCellTemplateExample periodCellTemplateExample1 = new PeriodCellTemplateExample();
//                        periodCellTemplateExample1.createCriteria().andCellTemplateIdEqualTo(keyInCellTemplateConfig.getCellTemplateId());
//                        Optional<PeriodCellTemplate> periodCellTemplate = periodCellTemplateMapper.selectByExample(periodCellTemplateExample1).stream().findFirst();
//
//
//                        DataSource dataSource = new DataSource();
//                        dataSource.setId(distributedIdService.nextId());
//                        if (periodCellTemplate.isPresent()) {
//                            dataSource.setColumnIndex(periodCellTemplate.get().getColumnIndex());
//                            dataSource.setRowIndex(periodCellTemplate.get().getRowIndex());
//                        }
//                        dataSource.setAmount(new BigDecimal("0"));
//                        dataSource.setName("ManualDataSource");
//                        dataSource.setDescription("ManualDataSource");
//                        dataSource.setCreateTime(createTime);
//                        dataSource.setUpdateTime(createTime);
//                        dataSource.setCreateBy("Admin");
//                        dataSource.setUpdateBy("Admin");
//                        dataSource.setPeriod(period);
//                        dataSource.setCellTemplateId(keyInCellTemplateConfig.getCellTemplateId());
//                        dataSource.setType(FormulaDataSourceType.KeyInSource.getCode());
//                        dataSourceMapper.insertSelective(dataSource);
//                    }

                    //update formulablock table reportid field
                    List<Long> cellTemplateConfigIds = periodCellTemplateConfigs.stream()
                            .map(PeriodCellTemplateConfig::getCellTemplateId)
                            .collect(Collectors.toList());
                    if (cellTemplateConfigIds.size() > 0) {
                        periodFormulaBlockMapper.updateReportId(reportId, cellTemplateConfigIds, period, projectId);
                    }

                    for (PeriodCellTemplateConfig periodCellTemplateConfig : periodCellTemplateConfigs) {
                        PeriodFormulaBlockExample periodFormulaBlockExample2 = new PeriodFormulaBlockExample();
                        periodFormulaBlockExample2.createCriteria()
                                .andProjectIdEqualTo(projectId)
                                .andCellTemplateIdEqualTo(periodCellTemplateConfig.getCellTemplateId())
                                .andReportIdEqualTo(reportId)
                                .andPeriodEqualTo(period);
                        List<PeriodFormulaBlock> periodFormulaBlocks = periodFormulaBlockMapper.selectByExample(periodFormulaBlockExample2);

                        //TODO：如果formula 为 ND(100) +ND(22) ,需要使用正则表达式拆分出自定义公式，然后根据自定义公式取formulablock 的数据进行替换
                        String regex = "[A-Z]*\\([\\-A-Za-z0-9\\\"\\,\\.\\u4e00-\\u9fa5\\%]*\\)";
                        Pattern p = Pattern.compile(regex);
                        String sourceFormula = StringUtils.isNotBlank(periodCellTemplateConfig.getKeyValueParsedFormula()) ?
                                periodCellTemplateConfig.getKeyValueParsedFormula() :
                                periodCellTemplateConfig.getFormula();
                        String resultFormula = StringUtils.isNotBlank(periodCellTemplateConfig.getKeyValueParsedFormula()) ?
                                periodCellTemplateConfig.getKeyValueParsedFormula() :
                                periodCellTemplateConfig.getFormula();
                        Matcher m = p.matcher(sourceFormula);
                        Boolean isFind = false;
                        while (m.find()) {
                            isFind = true;
                            //如果有些公式无法用正则匹配，可以做特殊处理
                            String findStr = m.group();
                            Optional<PeriodFormulaBlock> formulaBlock = periodFormulaBlocks.stream()
                                    .filter(a -> a.getFormulaExpression().equals(findStr))
                                    .findFirst();
                            if (formulaBlock.isPresent()) {
                                resultFormula = resultFormula.replace(findStr, formulaBlock.get().getData());
                            }
                        }

                        //如果有正则匹配就进行更新公式解析
//                        if (isFind) {
                        periodCellTemplateConfig.setParsedFormula(resultFormula);
                        periodCellTemplateConfigMapper.updateByPrimaryKeySelective(periodCellTemplateConfig);
//                        }

                        String regexNormalCell = "[A-Z]{1,2}((?!0)[0-9]{1,3})";
                        p = Pattern.compile(regexNormalCell);
                        m = p.matcher(sourceFormula);
                        while (m.find()) {
                            //如果找到普通单元格的公式，就去workbook里查找对应的格子取数据，然后放到DataSource，作为普通单元格数据源的数据
                            //找到一个格子就加一条数据
                            String findStr = m.group();//A12,A13,A15,how to get A and 12 or A and 13
                            String regexColumn = "[A-Z]{1,2}";
                            Pattern pp = Pattern.compile(regexColumn);
                            Matcher mm = pp.matcher(findStr);
                            while (mm.find()) {
                                String colStr = mm.group();
                                int colNum = FormulaHelper.excelColStrToNum(colStr, colStr.length());
                                int rowNum = Integer.parseInt(StringUtils.removeStart(findStr, colStr));
                                Row row = sheet.getRow(rowNum - 1);
                                if (row != null) {
                                    Cell cell = row.getCell(colNum - 1);
                                    if (cell != null) {
                                        //开始取值然后存放到DataSource
                                        PeriodDataSource dataSource = new PeriodDataSource();
                                        dataSource.setId(distributedIdService.nextId());
                                        dataSource.setColumnIndex(colNum - 1);
                                        dataSource.setRowIndex(rowNum - 1);
                                        if (((XSSFCell) cell).getRawValue() != "#VALUE!") {
                                            dataSource.setAmount(new BigDecimal(
                                                    ((XSSFCell) cell).getRawValue() != null ?
                                                            ((XSSFCell) cell).getRawValue()
                                                            : (StringUtils.isNotBlank(cell.getStringCellValue())) ?
                                                            cell.getStringCellValue()
                                                            : Double.toString(cell.getNumericCellValue())));
                                        } else {
                                            dataSource.setAmount(new BigDecimal("0.0"));
                                        }
                                        dataSource.setName("ReportDataSource");
                                        dataSource.setDescription(findStr);
                                        dataSource.setCreateTime(createTime);
                                        dataSource.setUpdateTime(createTime);
                                        dataSource.setCreateBy("Admin");
                                        dataSource.setUpdateBy("Admin");
                                        dataSource.setPeriod(period);
                                        dataSource.setCellTemplateId(periodCellTemplateConfig.getCellTemplateId());
                                        dataSource.setType(FormulaDataSourceType.Report.getCode());
                                        periodDataSourceMapper.insertSelective(dataSource);
                                        //这里有个问题就是DataSource的数据有了，但是celldatasource的数据没有，后面无法关联celldata和DataSource
                                        //解决办法就是 在存DataSource的时候就先把celldata的数据加好
                                        //然后把celldatasource的数据也加好
                                        //然后在外面去更新celldata和celldatasource的数据
                                        //还有就是把公式里的数据源放到内存，在所有算完之后再统一加数据源
                                    }
                                }
                            }
                        }

                        Optional<PeriodCellTemplate> tempPeriodCellTemplate = periodCellTemplateList.stream()
                                .filter(a -> a.getCellTemplateId().equals(periodCellTemplateConfig.getCellTemplateId()))
                                .findFirst();
                        if (tempPeriodCellTemplate.isPresent()) {
                            PeriodCellData cellData = new PeriodCellData();
                            Long cellDataId = distributedIdService.nextId();
                            cellData.setId(cellDataId);
                            cellData.setReportId(reportId);
                            cellData.setCellTemplateId(tempPeriodCellTemplate.get().getCellTemplateId());
                            String data;
                            if (sheet.getRow(tempPeriodCellTemplate.get().getRowIndex()) != null
                                    && sheet.getRow(tempPeriodCellTemplate.get().getRowIndex())
                                    .getCell(tempPeriodCellTemplate.get().getColumnIndex()) != null) {
                                Cell cell = sheet.getRow(tempPeriodCellTemplate.get().getRowIndex())
                                        .getCell(tempPeriodCellTemplate.get().getColumnIndex());
                                data = ((XSSFCell) cell).getRawValue();
                                if (data != null && data.equals("#VALUE!")) {
                                    data = "0.0";
                                }
                                //evaluator.evaluate(cell)
//                                if (cell.getCellTypeEnum().equals(CellType.NUMERIC)||cell.getCellTypeEnum().equals(CellType.FORMULA)) {
//                                    data = Double.toString(cell.getNumericCellValue());
//                                } else {
//                                    data = cell.getStringCellValue();
//                                }
                            } else {
                                data = EMPTY;
                            }

                            if (StringUtils.isNotBlank(data)) {
                                Pattern pattern = Pattern.compile("[0-9.]*");
                                Matcher isNum = pattern.matcher(data);
                                if (isNum.matches()) {
                                    cellData.setData(new BigDecimal(data).toString());
                                } else {
                                    cellData.setData(data);
                                }
                            } else {
                                cellData.setData(data);
                            }

                            //cellData.setData(new BigDecimal(data).toString());
//                            PeriodFormulaBlockExample periodFormulaBlockExample = new PeriodFormulaBlockExample();
//                            periodFormulaBlockExample.createCriteria().andPeriodEqualTo(period)
//                                    .andCellTemplateIdEqualTo(tempPeriodCellTemplate.get().getCellTemplateId());

//                            if (isFind) {
                            cellData.setFormulaExp(resultFormula);
//                            } else {
//                                cellData.setFormulaExp(EMPTY);
//                            }
                            cellData.setCreateBy("Admin");
                            cellData.setCreateTime(createTime);
                            cellData.setUpdateBy("Admin");
                            cellData.setUpdateTime(createTime);
                            cellData.setProjectId(projectId);
                            cellData.setPeriod(period);
                            periodCellDataMapper.insertSelective(cellData);

                            //after insert celldata, insert the celldatasource for link celldata and datasource
                            PeriodDataSourceExample dataSourceExample = new PeriodDataSourceExample();
                            dataSourceExample.createCriteria().andPeriodEqualTo(period)
                                    .andCellTemplateIdEqualTo(tempPeriodCellTemplate.get().getCellTemplateId());
                            List<PeriodDataSource> dataSourceList = SpringContextUtil.periodDataSourceMapper.selectByExample(dataSourceExample);
                            for (int ii = 0; ii < dataSourceList.size(); ii++) {
                                PeriodDataSource dataSource = dataSourceList.get(ii);
                                PeriodCellDataSource cellDataSource = new PeriodCellDataSource();
                                cellDataSource.setId(distributedIdService.nextId());
                                cellDataSource.setCellTemplateId(tempPeriodCellTemplate.get().getCellTemplateId());
                                cellDataSource.setCellDataId(cellDataId);
                                cellDataSource.setDataSourceId(dataSource.getId());
                                cellDataSource.setCreateTime(createTime);
                                cellDataSource.setUpdateTime(createTime);
                                cellDataSource.setPeriod(period);
                                cellDataSource.setProjectId(projectId);
                                SpringContextUtil.periodCellDataSourceMapper.insertSelective(cellDataSource);
                            }
                        }
                    }
                }
            }
            return "generate report successful";
        } else {
            return "GenerateReport failed";
        }
    }

    private String getTemplatePath(String path) {
        String result;
        String PREFIX_VALUE = "~";
        if (path.startsWith(PREFIX_VALUE)) {
            result = path.substring(1, path.length());
        } else {
            result = path;
        }
        return result;
    }

    /**
     * 根据传入的模板code和模板路径创建workbook
     *
     * @param templates 模板code和模板路径 键值对
     * @return 工作簿workbook
     */
    private Optional<Workbook> createWorkBookWithExcelFileList(List<PeriodTemplate> templates) {
        Workbook workbook = new XSSFWorkbook();
        try {
            String filePath = this.getClass().getResource("").toURI().getPath();
            String tempPath = filePath.substring(0, filePath.indexOf("classes") + "\\classes".length());
            templates.forEach(a -> {
                Workbook tWorkbook = null;
                File file = null;
                if (a.getIsSystemType()) {
                    file = new File(tempPath + a.getPath());
                    try {
                        tWorkbook = WorkbookFactory.create(file);
                    } catch (IOException e) {
                        e.printStackTrace();
                    } catch (InvalidFormatException e) {
                        e.printStackTrace();
                    }
                } else {
                    InputStream is = null;
                    try {
                        is = ftpService.getFtpFileWithStaticUrl(a.getPath());
                        tWorkbook = WorkbookFactory.create(is);
                    } catch (Exception e) {
                        e.printStackTrace();
                    } finally {
                        if (is != null) {
                            try {
                                is.close();
                            } catch (IOException e) {
                                e.printStackTrace();
                            }
                        }
                    }
                }
                POIUtil.cloneSheet(tWorkbook.getSheetAt(0), workbook.createSheet(a.getCode()));
            });
            return Optional.of(workbook);
        } catch (Exception e) {
            logger.error("createWorkBookWithExcelFileList error.", e);
        }
        return Optional.empty();
    }

    /**
     * 注册所有的自定义方法到工作簿
     *
     * @param workbook 工作簿
     */
    private void addFunctionsToWorkbook(Workbook workbook, FormulaContext formulaContext) {
        String[] functionNames = {"SGSR", "FSJZ", "ND", "BB", "XXFP", "GZSD", "ProjectContext", "JXFPMX", "JXFP"};
        FreeRefFunction[] functionImpls = {new SGSR(formulaContext), new FSJZ(formulaContext), new ND(formulaContext),
                new BB(formulaContext), new XXFP(formulaContext), new GZSD(formulaContext), new ProjectContext(formulaContext)
                , new JXFPMX(formulaContext), new JXFP(formulaContext)};
        UDFFinder udfs = new DefaultUDFFinder(functionNames, functionImpls);
        UDFFinder udfToolpack = new AggregatingUDFFinder(udfs);
        workbook.addToolPack(udfToolpack);
    }

    public List<CellCalcInfoDto> getCellCalcInfo(List<Long> templateIdList, Integer periodParam, String projectId) {
        //already finished by 20180711
        int period = 0;
        if (periodParam != null) {
            period = periodParam;
        }

        List<PeriodCellTemplateConfigExtendDto> periodCellTemplateConfigExtendDtos =
                periodCellTemplateConfigMapper.getPeriodCellTemplateConfigExtendDtos(templateIdList, period, projectId);

        fixedPCTParsedFormula(periodCellTemplateConfigExtendDtos, projectId, period);
        List<CellCalcInfoDto> cellCalcInfoDtos = new ArrayList<>();
        periodCellTemplateConfigExtendDtos.stream().collect(Collectors.groupingBy(a ->
                new CellTemplateConfigGroupDto(a.getColumnIndex(), a.getRowIndex()
                        , a.getColumnName(), a.getRowName(), a.getDataType(), a.getReadOnly()
                        , a.getCellTemplateId(), a.getReportTemplateId(), a.getComment()).toStr())).forEach((x, y) -> {
            Optional<Map<String, Object>> result = x.values().stream().findFirst();
            if (result.isPresent()) {
                CellCalcInfoDto cellCalcInfoDto = new CellCalcInfoDto();
                cellCalcInfoDto.setColumnIndex(Integer.parseInt(result.get().get("columnIndex").toString()));
                cellCalcInfoDto.setColumnName(result.get().get("columnName").toString());
                cellCalcInfoDto.setRowIndex(Integer.parseInt(result.get().get("rowIndex").toString()));
                cellCalcInfoDto.setRowName(result.get().get("rowName").toString());
                cellCalcInfoDto.setFormula(convertListToString(y.stream()
                        .map(PeriodCellTemplateConfig::getFormula).collect(Collectors.toList())));
                cellCalcInfoDto.setParsedFormula(convertListToString(y.stream()
                        .map(PeriodCellTemplateConfig::getParsedFormula).collect(Collectors.toList())));
                //todo: add validation at here later
                cellCalcInfoDto.setCellTemplateId(result.get().get("cellTemplateId").toString());
                cellCalcInfoDto.setReportTemplateId(Long.parseLong(result.get().get("reportTemplateId").toString()));
                cellCalcInfoDto.setDataType(Integer.parseInt(result.get().get("dataType").toString()));
                cellCalcInfoDto.setIsReadOnly(Boolean.valueOf(result.get().get("isReadOnly").toString()));
                cellCalcInfoDto.setComment(result.get().get("comment").toString());
                cellCalcInfoDto.setConfigList(y.stream()
                        .map(PeriodCellTemplateConfigExtendDto::getPeriodCellTemplateConfig)
                        .collect(Collectors.toList()));
                cellCalcInfoDtos.add(cellCalcInfoDto);
            }
        });
        return cellCalcInfoDtos;
    }

    private void fixedPCTParsedFormula(List<PeriodCellTemplateConfigExtendDto> periodCellTemplateConfigExtendDtos, String projectId, int period) {
        Map<String, List<PCTEntity>> formulaMapToPCT = new HashMap<>();
        Map<PeriodCellTemplateConfig, List<String>> configMapToPCTs = new HashMap<>();
        Set<PCTEntity> parameter = new HashSet<>();

        periodCellTemplateConfigExtendDtos.forEach(p -> {
            String formula = p.getFormula();
            String parsedFormula = p.getParsedFormula();
            if (formula.contains("BB(")) {
                logger.debug("period cell template config contains bb formula {}", formula);
                if (parsedFormula.contains("PCT(")) {
                    List<String> parsedPCTs = new ArrayList<>();
                    byte[] parsedFormulaBytes = parsedFormula.getBytes();
                    Integer begin = null;

                    for (int i = 0; i < parsedFormulaBytes.length; i++) {
                        if (i < parsedFormulaBytes.length - 3 && parsedFormulaBytes[i] == 'P' && parsedFormulaBytes[i + 1] == 'C'
                                && parsedFormulaBytes[i + 2] == 'T' && parsedFormulaBytes[i + 3] == '(') {
                            begin = i;
                        } else if (parsedFormulaBytes[i] == ')' && begin != null) {
                            parsedPCTs.add(new String(Arrays.copyOfRange(parsedFormulaBytes, begin, i + 1)));
                            begin = null;
                        }
                    }
                    if (!parsedPCTs.isEmpty()) {
                        configMapToPCTs.put(p, parsedPCTs);
                    }

                    parsedPCTs.stream().forEach(m -> {
                        String periodCPT = m.substring(m.indexOf("(") + 1, m.indexOf(")"));
                        if (periodCPT.contains(",")) {
                            formulaMapToPCT.put(m, Arrays.asList(periodCPT.split(",")).stream().map(n -> new PCTEntity(n))
                                    .collect(Collectors.toList()));
                        } else {
                            formulaMapToPCT.put(m, Lists.newArrayList(new PCTEntity(periodCPT)));
                        }
                        parameter.addAll(formulaMapToPCT.get(m));
                    });


                } else {
                    logger.warn("bb formula parsedformula must contains PCT but not found {}", parsedFormula);
                }
            }
        });

        if (!parameter.isEmpty()) {
            List<PCTEntity> pctResults = periodCellDataMapper.queryByPCTs(parameter, projectId);
            Map<PCTEntity, BigDecimal> pctCache = new HashMap<>();
            pctResults.forEach(m -> {
                BigDecimal data = null;
                try {
                    data = new BigDecimal(m.getData());
                } catch (NumberFormatException e) {
                    logger.warn("number format ecption for parameter {}", m);
                    data = BigDecimal.ZERO;
                }
                pctCache.put(m, data);
            });

            configMapToPCTs.forEach((k, v) -> {
                v.forEach(pctStr -> {
                    List<PCTEntity> entities = formulaMapToPCT.get(pctStr);
                    BigDecimal result = new BigDecimal("0");
                    for (PCTEntity entity : entities) {
                        result = pctCache.containsKey(entity) ? result.add(pctCache.get(entity)) : result;
                    }
                    k.setParsedFormula(k.getParsedFormula().replace(pctStr, result.toString()));
                });

            });
        }
    }

    private String convertListToString(List<String> list) {
        StringBuilder stringBuilder = new StringBuilder();
        list.forEach(s -> {
            if (StringUtils.isNotBlank(s)) {
                stringBuilder.append(s);
                stringBuilder.append("+");
            }
        });
        if (stringBuilder.length() > 0) {
            stringBuilder.deleteCharAt(stringBuilder.length() - 1);
        }
        return stringBuilder.toString();
    }


}
